/* brick.c */

#include <stdio.h>

#include "mur.h"		/* global definitions */
#include "interf.h"
#include "scrio.h"
#include "mem.h"
#include "brick.h"
#include "engio.h"

#include "externs.h"



Brick *brick_list;				       /* doubly linked list */


Brick *current_brick;

static Brick *bricks_freelist = NULL;
Brick *bricks_touched = NULL;
static Brick *bricks_modified = NULL;

static char spaces_line [X_BRICK+1];
static char holes_line [X_BRICK+1];


void init_bricks ()
{
    static int initialized = 0;

    if (!initialized) {
	initialized = 1;
	brick_list = New (Brick);		     /* freelist is empty... */
	brick_list->next = brick_list;
	brick_list->prev = brick_list;
	memset (spaces_line, ' ', X_BRICK);
	memset (holes_line, 127, X_BRICK);
	spaces_line [X_BRICK] = '\0';
	holes_line [X_BRICK] = '\0';
    }
}

/* ********************************************************************** */


static void free_brick (brick)
    Brick *brick;
{
    if (brick == NULL)
	return;
    brick->next->prev = brick->prev;
    brick->prev->next = brick->next;
    brick->next = bricks_freelist;
    bricks_freelist = brick;
}


void free_all_bricks ()
{
    while (brick_list->next != brick_list)
	free_brick (brick_list->next);
    bricks_touched = NULL;
    bricks_modified = NULL;
}


Brick *new_brick ()
{
    Brick *brick;

    if (bricks_freelist != NULL) {
	brick = bricks_freelist;
	bricks_freelist = bricks_freelist->next;
    } else
	brick = New (Brick);
    /* Add to doubly linked list. */
    brick->next = brick_list->next;
    brick->prev = brick_list;
    brick->next->prev = brick;
    brick_list->next = brick;
    /* Mark touched, and add to touched list. */
    /* The caller has to set itself the touched_line [] */
    brick->touched = 1;
    brick->next_touched = bricks_touched;
    bricks_touched = brick;
    brick->modified = 0;
    brick->flags = 0;
    return brick;
}



Brick *search_brick (x, y)					 /* char coo */
    long x, y;
{
    Brick *br;

    x = Abs_brick_x (x);
    y = Abs_brick_y (y);
    for (br = brick_list->next; br != brick_list; br = br->next)
	if (br->x == x && br->y == y)
	    return br;
    return NULL;
}



void mark_touchedline_brick (brick, dy)
    Brick *brick;
    long dy;
{
    brick->touched_line [dy] = 1;
    if (! brick->modified) {
	brick->modified = 1;
	brick->next_modified = bricks_modified;
	bricks_modified = brick;
    }
    if (! brick->touched) {
	brick->touched = 1;
	brick->next_touched = bricks_touched;
	bricks_touched = brick;
    }
}

#if 0
void update_empty_brick (brick)					 /* UNUSED ? */
    Brick *brick;
{
    long i;
    for (i=0; i<Y_BRICK; i++)
	if (! brick->empty_line [i]) {
	    brick->flags &= ~BR_EMPTY;
	    return;
	}
    brick->flags |= BR_EMPTY;
}
#endif


#if 0
void update_emptyline_brick (brick, dy, xbeg, nb)		 /* UNUSED ? */
    Brick *brick;
    long dy;
    long xbeg, nb;
{
    int i;
    char *line;
    line = brick->line [dy];
    if (brick->empty_line [dy]) {
	for (i=xbeg; i < xbeg+nb; i++)
	    if (line [i] != ' ') {
		brick->empty_line [dy] = 0;
		return;
	    }
    } else {
	for (i=0; i<X_BRICK; i++)
	    if (line [i] != ' ')
		return;
	brick->empty_line [dy] = 1;
    }
    return;
}
#endif




static char *resize_line (str, len)
    char *str;
    long len;
{
    long cnt;

    cnt = strlen (str);
    while (cnt < len)
	str [cnt++] = ' ';
    str [len] = '\0';
    return str;
}



static void read_a_brick (br)
    Brick *br;
{
    strmax recvd;
    char *ptr;
    long y;
    int empty_brick, empty_line;
    int er;

    read_server (recvd);
    if (!strcmp (EMPT, recvd)) {			      /* EMPTY brick */
	for (y=0; y<Y_BRICK; y++) {
	    memcpy (br->line [y], spaces_line, X_BRICK+1);
	    memcpy (br->orig [y], spaces_line, X_BRICK+1);
	    br->empty_line [y] = 1;
	    br->touched_line [y] = 1;
	}
	br->flags |= BR_EMPTY;
	return;
    }
    if (!strcmp (BRIK, recvd)) {
	empty_brick = 1;
	for (y=0; y<Y_BRICK; y++) {
	    read_server (recvd);
	    memcpy (br->orig [y], resize_line (recvd, X_BRICK), X_BRICK+1);
	    memcpy (br->line [y], br->orig [y], X_BRICK+1);
	    empty_line = 1;
	    for (ptr=recvd; *ptr; ptr++)
		if (*ptr != ' ') {
		    empty_line = 0;
		    break;
		}
	    br->empty_line [y] = empty_line;
	    br->touched_line [y] = 1;
	    empty_brick &= empty_line;
	}
	if (empty_brick)
	    br->flags |= BR_EMPTY;
	read_server (recvd);
	if (!strcmp (FBRK, recvd)) {
	    return;
	}
	if ((er = get_errno (recvd)) != 0) {
	error:
	    strncpy (br->orig [0],
		     "Bad Brick !^^^^^^^^^^^^^^^^^^^^\\", X_BRICK);
	    br->orig [1][0] = '|';
	    strncpy (br->orig [1]+1, recvd, X_BRICK-2);	/* error message */
	    resize_line (br->orig [1], X_BRICK);
	    br->orig [1][X_BRICK-1] = '|';
	    for (y=2; y<Y_BRICK-1; y++) {
		strcpy (br->orig [y], spaces_line);
		br->orig [y][0] = '|';
		br->orig [y][X_BRICK-1] = '|';
	    }
	    strncpy (br->orig [Y_BRICK-1],
		     "\\______________________________/", X_BRICK);
	    for (y=0; y<Y_BRICK; y++) {
		strcpy (br->line [y], br->orig [y]);
		br->empty_line [y] = 0;
		br->touched_line [y] = 1;
	    }
	    br->flags |= BR_BAD;
	    return;
	}
	panic ("Read brick, bad terminator : %s", recvd);
    }
    if ((er = get_errno (recvd)) != 0)
	goto error;
    panic ("Read brick, unknown : %s", recvd);
}



static Brick *force_get_brick (x, y, br)
    long x, y;
    Brick *br;
{
    strmax tmp;

    x = Abs_brick_x (x);
    y = Abs_brick_y (y);
    br->x = x;
    br->y = y;
    br->flags = 0;
    if (! br->touched) {
	br->touched = 1;
	br->next_touched = bricks_touched;
	bricks_touched = br;
    }
    if (br->modified) {	/* inefficient */
	Brick *b;
	Brick **bp = &bricks_modified;
	while ((b = *bp) != NULL) {
	    if (b == br) {
		*bp = b->next_modified;
		break;
	    }
	}
	br->modified = 0;
    }
    sprintf (tmp, "%s %ld %ld", READ, x, y);
    flush_write_server (tmp);
    read_a_brick (br);
    return br;
}


Brick *get_brick (x, y)
    long x, y;
{
    Brick *brick;

    x = Abs_brick_x (x);
    y = Abs_brick_y (y);
    brick = search_brick (x, y);
    if (brick != NULL)
	return brick;
    brick = new_brick ();
    return force_get_brick (x, y, brick);
}


void warp ()
{
    strmax recvd;

    flush_write_server (WARP);
    eread_server (recvd);
    if (!strcmp (WRPD, recvd))
	return;
    panic ("Guardian can't warp...");
}


static void flush_brick (br)
    Brick *br;
{
    strmax tmp;
    long i, x, y;

    if (br == NULL)
	return;
    if (br->flags & BR_BAD)
	return;
    sprintf (tmp, "%s %ld %ld", WRIT, br->x, br->y);
    write_server (tmp);
    for (y=0; y<Y_BRICK; y++) {
	char line [X_BRICK+1];
	char *ptr = line;
	int last_nonhole = -1;
	for (x=0; x<X_BRICK; x++) {
	    char c = br->line[y][x];
	    char *op = &(br->orig[y][x]);
	    if (*op == c)
		*ptr++ = 127;	/* hole */
	    else {
		*op = c;
		*ptr++ = c;
		last_nonhole = x;
	    }
	}
	line [last_nonhole+1] = '\0';
	write_server (line);
    }
    flush_write_server (FBRK);
    
    read_server (tmp);
    if (strcmp (WRTN, tmp)) {
	i = get_errno (tmp);
	if (i != 8) { /* not "read only brick" */
	    panic ("Flush brick, wrtn expected");
	}
    }
}


void flush_all_bricks ()				 /* may stay touched */
{
    Brick *brick;

    for (brick = bricks_modified; brick != NULL; brick = brick->next_modified){
	flush_brick (brick);
	brick->modified = 0;
    }
    bricks_modified = NULL;
}








static Queue *the_queue = NULL;

/*
 * freelist for Queue
 */

static Queue *Queue_freelist = NULL;

static Queue *new_Queue ()
{
    Queue *q;
    if (Queue_freelist) {
	q = Queue_freelist;
	Queue_freelist = q->next;
    } else
	q = New (Queue);
    return q;
}

static void free_Queue (q)
    Queue *q;
{
    q->next = Queue_freelist;
    Queue_freelist = q;
}



static void add_brick_queue (x, y) /* must keep the order !!! */
    long x, y;
{
    Queue *new, *qu, **qptr;

    x = Abs_brick_x (x);
    y = Abs_brick_y (y);
    qptr = &the_queue;
    while ((qu = *qptr) != NULL) {
	if (qu->x == x && qu->y == y)
	    return;
	qptr = &(qu->next);
    }
    new = new_Queue ();
    new->next = NULL;
    new->x = x;
    new->y = y;
    *qptr = new;
}


void add_burst_zone (bx, by, nbx, nby)
    long bx, by, nbx, nby;
{
    long i, j;
    for (j=0; j<nby; j++)
	for (i=0; i<nbx; i++)
	    add_brick_queue (bx+i, by+j);
}


static void clear_brick_queue ()
{
    while (the_queue != NULL) {
	Queue *next = the_queue->next;
	free_Queue (the_queue);
	the_queue = next;
    }
}


void read_brick_queue ()
{
    Queue *qu, **qptr;
    Brick *br;
    strmax tmp;
    long x, y;

    /*
     * Ignore the bricks that are already in memory.
     */
    qptr = &the_queue;
    while ((qu = *qptr) != NULL) {
	if (search_brick (qu->x, qu->y) != NULL) {
	    *qptr = qu->next;
	    free_Queue (qu);
	    continue;
	}
	qptr = &(qu->next);
    }
    /*
     * Send requests for bricks.
     */
    for (qu = the_queue; qu != NULL; qu = qu->next) {
	x = qu->x;
	y = qu->y;
	sprintf (tmp, "%s %ld %ld", READ, x, y);
	write_server (tmp);
    }
    /*
     * Now, read the incoming bricks.
     */
    qu = the_queue;
    while (qu != NULL)
    {
	br = new_brick ();
	br->x = x = qu->x;
	br->y = y = qu->y;
	qu = qu->next;
	read_a_brick (br);
	flush_once ();
    }
    clear_brick_queue ();
}



#if 0
void actual_brick (x, y, dist)					 /* UNUSED ? */
    long x, y;
    long dist;
{
    int i, j;
    long px, py, bx, by, bxf, byf, dbx, dby;
    PBrick br = beg_brick, br2;
    bx = x >> X_SHIFT;
    by = y >> Y_SHIFT;
    dbx = Abs_brick_x (((x+cols-1) >> X_SHIFT) - bx);
    dby = Abs_brick_y (((y+lines-1) >> Y_SHIFT) - by);
    while (br)
    {
	px = br->x;
	py = br->y;
	br2 = br->next;
	if (Isout (px,py,bx,by,dbx,dby))
	{
	    flush_brick (br);
	    free_brick (br);
	}
	br = br2;
    }
}
#endif



void reask_bricks ()
{
    Brick *br;

    flush_all_bricks ();
    for (br = brick_list->next; br != brick_list; br = br->next)
	force_get_brick (br->x, br->y, br);
}


