/* builtin.c */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h> /* for fd_set */

#include "toplevel.h"
#include "scrio.h"

#include "commands.h"
#include "mem.h"
#include "list.h"
#include "error.h"
#include "util.h"
#include "parse.h"
#include "builtin.h"

#include "externs.h"

/* ************************************************ */
/* *************** Builtin commands *************** */
/* ************************************************ */


int cmd_nop ()
{
    return 0;
}



int cmd_eval (l)
    List *l;
{
    if (l == NULL || l->next != NULL) {
	set_error_list ("eval", l, "bad number of arguments");
	return 0;
    }
    exec_string (l->arg);
    return 0;
}
	    

int cmd_echo (l)
    List *l;
{
    int i;
    List *ll;
    char *ptr, *string;
    int lastnl = 1;

    while (l != NULL) {
	if (STREQ (l->arg, "--")) {
	    l = l->next;
	    goto doit;
	}
	if (STREQ (l->arg, "-n")) {
	    l = l->next;
	    lastnl = 0;
	    continue;
	}
	break;
    }
 doit:
    i = 0;
    for (ll=l; ll != NULL; ll = ll->next)
	i += strlen (ll->arg) + 1;
    ptr = string = mmalloc (i+1);  /* room for spaces and the last \n and \0 */
    for (; l != NULL; l = l->next) {
	int sz = strlen (l->arg);
	memcpy (ptr, l->arg, sz);
	ptr += sz;
	*ptr++ = ' ';
    }
    if (lastnl) {
	ptr [-1] = '\n';
	ptr++;
    }
    ptr [-1] = '\0';

    print_string (string);
    mfree (string);
    return 0;
}



int cmd_exec (l)
    List *l;
{
    int fd [2];
#define EXEC_BUF_SZ 1000    
    static char buffer[EXEC_BUF_SZ+1];

    if (l == NULL || l->next != NULL) {
	set_error_list ("exec", l, "bad number of arguments");
	return 0;
    }
    if (pipe (fd) != 0) {
	set_error_errno ("exec");
	return 0;
    }
    if (mode == MODE_SHELL)
	stty_original ();
    /* else ; */

    switch (fork ()) {
    case -1:
	set_fixed_error (ERR_EXEC, "couldn't fork");
	return 0;
    case 0: { /* child */
	int i;
	for (i=1; i<32; i++)
	    signal (i, SIG_DFL);
	close (fd [0]);
	fd [1] = fd_safe (fd [1]);
	dup2 (fd [1], 1);
	dup2 (fd [1], 2);
	close (fd [1]);
	execl ("/bin/sh", "sh", "-c", l->arg, NULL);
	exit (0);
	/*NOTREACHED*/
    }
    }
    close (fd [1]);
    while (1) {
	int i, j;
	int res = read (fd [0], buffer, EXEC_BUF_SZ);
	if (res < 0) {
	    extern int errno;
	    fprintf (stderr, "\nread ? errno=%d\n", errno);
	    break;
	}
	if (res == 0)
	    break;
	j = i = 0;
	while (i < res) {
	    char c;
	    if ((c = buffer[i++]) != '\0')			/* remove \0 */
		buffer[j++] = c;
	}
	if (j) {
	    buffer[j] = '\0';
	    print_string (buffer);
	}
    }
    stty_running ();
    close (fd [0]);
    return 0;
}


int cmd_pipe (l)
    List *l;
{
    int fdin [2];
    int fdout [2];
    char *string, *ptr;
    long sz;
    List *ll;
#define PIPE_BUF_SZ 1000    
    static char buffer[PIPE_BUF_SZ+1];
    fd_set fdsetin, fdsetout;
    fd_set *fdsetinp, *fdsetoutp;
    int nbfd;
    int inputnotclosed, outputnotclosed;
    int lastnl = 1;

    while (l != NULL) {
	if (STREQ (l->arg, "--")) {
	    l = l->next;
	    goto doit;
	}
	if (STREQ (l->arg, "-n")) {
	    l = l->next;
	    lastnl = 0;
	    continue;
	}
	break;
    }
 doit:
    if (l == NULL) {
	set_error_list ("pipe", l, "no arguments");
	return -1;
    }

    if (pipe (fdin) != 0 || pipe (fdout) != 0) {
	set_error_errno ("pipe");
	return -1;
    }
    switch (fork ()) {
    case -1:
	set_fixed_error (ERR_EXEC, "couldn't fork");
	return -1;
    case 0: { /* child */
	int i;
	for (i=1; i<32; i++)
	    signal (i, SIG_DFL);
	close (fdout [1]);
	close (fdin [0]);
	fdout [0] = fd_safe (fdout [0]);
	fdin [1] = fd_safe (fdin [1]);
	dup2 (fdout [0], 0);
	dup2 (fdin [1], 1);
	dup2 (fdin [1], 2);
	close (fdout [0]);
	close (fdin [1]);
	execl ("/bin/sh", "sh", "-c", l->arg, NULL);
	exit (0);
	/*NOTREACHED*/
    }
    }
    close (fdin [1]);
    close (fdout [0]);

    sz = 0;
    if (l->next != NULL) {
	for (ll=l->next; ll != NULL; ll = ll->next)
	    sz += strlen (ll->arg) + 1;
	ptr = string = mmalloc (sz+1);
	for (ll=l->next; ll != NULL; ll = ll->next) {
	    long i = strlen (ll->arg);
	    memcpy (ptr, ll->arg, i);
	    ptr += i;
	    *ptr++ = '\n';
	}
	if (lastnl) {
	    sz++;
	    ptr++;
	}
	ptr [-1] = '\0';
	sz--;
    } else
	string = NULL;

    fdsetinp = &fdsetin;
    fdsetoutp = &fdsetout;
    inputnotclosed = 1;
    if (sz) {
	ptr = string;
	outputnotclosed = 1;
	nbfd = ((fdin [0] > fdout [1]) ? fdin [0] : fdout [1]) + 1;
    } else {
	close (fdout [1]);
	outputnotclosed = 0;
	nbfd = fdin [0] + 1;
	fdsetoutp = NULL;
    }

    while (inputnotclosed || outputnotclosed) {
	int sel;
	if (inputnotclosed) {
	    FD_ZERO (fdsetinp);
	    FD_SET (fdin [0], fdsetinp);
	}
	if (outputnotclosed) {
	    FD_ZERO (fdsetoutp);
	    FD_SET (fdout [1], fdsetoutp);
	}
	sel = select (nbfd, fdsetinp, fdsetoutp, NULL, NULL);
	if (sel == -1)
	    continue; /* EINTR */
	if (inputnotclosed && FD_ISSET (fdin [0], fdsetinp)) {
	    int len;
	    len = read (fdin [0], buffer, PIPE_BUF_SZ);
	    if (len < 0) {
		extern int errno;
		fprintf (stderr, "\nread ? errno=%d\n", errno);
		break;
	    }
	    if (len == 0) {
		close (fdin [0]);
		inputnotclosed = 0;
		nbfd = fdout [1] + 1;
		fdsetinp = NULL;
	    } else {
		int i, j;
		j = i = 0;
		while (i < len) {
		    char c;
		    if ((c = buffer[i++]) != '\0')
			buffer[j++] = c;
		}
		if (j) {
		    buffer[j] = '\0';
		    print_string (buffer);
		}
	    }
	}
	if (outputnotclosed && FD_ISSET (fdout [1], fdsetoutp)) {
	    int len;
	    len = write (fdout [1], ptr, sz);
	    if (len <= 0) {
		if (len < 0 && errno != EWOULDBLOCK && errno == EAGAIN) {
		    /* if EPIPE */
		    close (fdout [1]);
		    outputnotclosed = 0;
		    nbfd = fdin [0] + 1;
		    fdsetoutp = NULL;
		} else {
		    /* len == 0 || errno == EWOULDBLOCK || errno == EAGAIN */
		    /* Shouldn't happen, but it doesn't hurt... */
		}
	    } else {
		ptr += len;
		sz -= len;
		if (sz == 0) {
		    close (fdout [1]);
		    outputnotclosed = 0;
		    nbfd = fdin [0] + 1;
		    fdsetoutp = NULL;
		}
	    }
	}
    }
    if (string != NULL)
	mfree (string);
    return 0;
}



int cmd_source (l)
    List *l;
{
    FILE *file;

    if (l == NULL || l->next != NULL) {
	set_error_list ("source", l, "bad number of arguments");
	return 0;
    }
    file = fopen (l->arg, "r");
    if (file == NULL) {
	set_error_list_errno ("source", l);
	return 0;
    }
    exec_file (file);
    return 0;
}


int cmd_cd (l)
    List *l;
{
    if (l == NULL || l->next != NULL) {
	set_error_list ("cd", l, "bad number of arguments");
	return 0;
    }
    if (chdir (l->arg)) {
	set_error_list_errno ("cd", l);
	return 0;
    }
    return 0;
}
   

int cmd_explode (l)
    List *l;
{
    long len;
    long i;
    char *mem, *ptr, *arg;
    if (l == NULL || l->next != NULL) {
	set_error_list ("explode", l, "bad number of arguments");
	return 0;
    }
    arg = l->arg;
    len = strlen (arg);
    ptr = mem = mmalloc (2*len+1);
    for (i=0; i<len; i++) {
	*ptr++ = *arg++;
	*ptr++ = '\n';
    }
    *ptr = '\0';
    print_string (mem);
    mfree (mem);
    return 0;
}

int cmd_implode (l)
    List *l;
{
    List *ll;
    long sz;
    char *mem, *ptr;
    char *sep;
    long seplen;
    int lastsep = 1;

    while (l != NULL) {
	if (STREQ (l->arg, "--")) {
	    l = l->next;
	    goto doit;
	}
	if (STREQ (l->arg, "-n")) {
	    l = l->next;
	    lastsep = 0;
	    continue;
	}
	break;
    }
 doit:
    if (l == NULL) {
	set_fixed_error (ERR_EXEC, "implode: bad number of arguments");
	return 0;
    }
    sep = l->arg;
    seplen = strlen (sep);
    sz = 0;
    for (ll=l->next; ll != NULL; ll = ll->next)
	sz += strlen (ll->arg) + seplen;
    ptr = mem = mmalloc (sz+1);  /* room for the last \0 */
    for (ll=l->next; ll != NULL; ll = ll->next) {
	long i = strlen (ll->arg);
	memcpy (ptr, ll->arg, i);
	ptr += i;
	if (seplen) {
	    memcpy (ptr, sep, seplen);
	    ptr += seplen;
	}
    }
    if (!lastsep)
	ptr -= seplen;
    *ptr = '\0';
    print_string (mem);
    mfree (mem);
    return 0;
}


int cmd_shift (l)
    List *l;
{
    long nb;
    long k;
    long sz;
    char *mem, *ptr;
    List *ll;

    if (l == NULL) {
	set_fixed_error (ERR_EXEC, "shift: bad number of arguments");
	return 0;
    }
    if (long_from_string (l->arg, &nb)) {
	set_error_list ("shift", l, "bad value");
	return 0;
    }
    k = nb;
    sz = 0;
    if (k < 0) {
	sz = -k;
	k = 0;
    }
    ll = l->next;
    for (; k && ll != NULL; k--)
	ll = ll->next;
    for (; ll != NULL; ll = ll->next)
	sz += strlen (ll->arg) + 1;
    ptr = mem = mmalloc (sz+1);

    k = nb;
    if (k < 0) {
	for (k = -k; k; k--)
	    *ptr++ = '\n';
    }
    ll = l->next;
    for (; k && ll != NULL; k--)
	ll = ll->next;
    for (; ll != NULL; ll = ll->next) {
	long i = strlen (ll->arg);
	memcpy (ptr, ll->arg, i);
	ptr += i;
	*ptr++ = '\n';
    }
    *ptr = '\0';
    print_string (mem);
    mfree (mem);
    return 0;
}


int cmd_interval (l)
    List *l;
{
    long nb1, nb2;
    long i;
    long sz;
    char *mem, *ptr;

    if (l == NULL || l->next == NULL || l->next->next != NULL) {
	set_error_list ("interval", l, "bad number of arguments");
	return 0;
    }
    if (long_from_string (l->arg, &nb1)
	|| long_from_string (l->next->arg, &nb2)) {
	set_error_list ("interval", l, "bad value");
	return 0;
    }
    if (nb2 < nb1)
	return 0;
    sz = 0;
    for (i=nb1; i<=nb2; i++)
	sz += strlen (string_from_long (i)) + 1;
    ptr = mem = mmalloc (sz+1);
    for (i=nb1; i<=nb2; i++) {
	char *str = string_from_long (i);
	int len = strlen (str);
	memcpy (ptr, str, len);
	ptr += len;
	*ptr++ = '\n';
    }
    *ptr = '\0';
    print_string (mem);
    mfree (mem);
    return 0;
}


int cmd_substring (l)
    List *l;
{
    long pos, len;
    long sz;
    char *mem;
    char *arg;

    if (l == NULL || l->next == NULL || l->next->next == NULL
	|| l->next->next->next != NULL) {
	set_error_list ("substring", l, "bad number of arguments");
	return 0;
    }
    if (long_from_string (l->arg, &pos)
	|| long_from_string (l->next->arg, &len)) {
	set_error_list ("substring", l, "bad value");
	return 0;
    }
    arg = l->next->next->arg;
    sz = strlen (arg);
    if (pos < 0) {
	pos = sz - len;
	if (pos < 0) {
	    pos = 0;
	    len = sz;
	}
    } else {
	if (pos > sz)
	    pos = sz;
	if (pos + len > sz)
	    len = sz - pos;
    }
    mem = mmalloc (len+1);
    memcpy (mem, arg+pos, len);
    mem[len] = '\0';
    print_string (mem);
    mfree (mem);
    return 0;
}


int cmd_asc (l)
    List *l;
{
    unsigned char c;

    if (l == NULL || l->next != NULL) {
	set_error_list ("asc", l, "bad number of arguments");
	return 0;
    }
    c = l->arg[0];
    if (c == '\0') {
	set_fixed_error (ERR_EXEC, "asc: empty argument");
	return 0;
    }
    print_string (string_from_long ((long) c));
    return 0;
}

int cmd_chr (l)
    List *l;
{
    long nb;
    unsigned char c[2];

    if (l == NULL || l->next != NULL) {
	set_error_list ("chr", l, "bad number of arguments");
	return 0;
    }
    if (long_from_string (l->arg, &nb)
	|| nb <= 0
	|| nb >= 256) {
	set_error_list ("chr", l, "bad value");
	return 0;
    }
    c[0] = nb;
    c[1] = '\0';
    print_string (c);
    return 0;
}

int cmd_rnd (l)
    List *l;
{
    extern double drand48 ();
    long nb;

    if (l == NULL || l->next != NULL) {
	set_error_list ("rnd", l, "bad number of arguments");
	return 0;
    }
    if (long_from_string (l->arg, &nb)
	|| nb <= 0) {
	set_error_list ("rnd", l, "bad value");
	return 0;
    }
    print_string (string_from_long ((long)(drand48 () * nb)));
    return 0;
}

int cmd_strlen (l)
    List *l;
{
    long nb;

    if (l == NULL || l->next != NULL) {
	set_error_list ("strlen", l, "bad number of arguments");
	return 0;
    }
    nb = strlen (l->arg);
    print_string (string_from_long (nb));
    return 0;
}

int cmd_listlen (l)
    List *l;
{
    long nb;

    nb = nbargs_list (l);
    print_string (string_from_long (nb));
    return 0;
}

int cmd_raise (l)
    List *l;
{
    if (l == NULL || l->next != NULL) {
	set_error_list ("raise", l, "bad number of arguments");
	return 0;
    }
    set_error (ERR_RAISE, l->arg);
    return 0;
}

