/* commands.c */

#include <stdio.h>

#include "mem.h"
#include "tree.h"
#include "list.h"
#include "variable.h"
#include "error.h"
#include "parse.h"
#include "lex.h"
#include "funs.h"
#include "util.h"
#include "toplevel.h"
#include "scrio.h"	   /* Only for print_string.  Which should be moved. */
#include "commands.h"

#include "externs.h"

/*
 * After a tree is destroyed, the lists that were created from it
 * must never be referenced anymore.
 */

/*
 * I have decided to push the command_stack before
 * - a user function is executed (because it may call other functions, etc.)
 * - a primitive (fn, ...) that builds a list (or uses memory) is executed.
 */

Stack *command_stack;


static char *print_buffer = NULL;
static int print_buffer_max = 1000;
static int print_buffer_pos = 0;
static int directprint = 1;

static int inbraces = 0;
static List *local_variables = NULL;

static int break_num = 1;


int *default_ifs = NULL;				    /* int is faster */
static int *special_ifs = NULL;



void init_commands ()
{
    static int initialized = 0;

    if (!initialized) {
	initialized = 1;
	command_stack = create_stack ();
	init_parser ();
	init_function_table ();
	init_variable_table ();
	print_buffer = mmalloc (print_buffer_max);
	default_ifs = (int *) mmalloc (256 * sizeof (int));
	special_ifs = (int *) mmalloc (256 * sizeof (int));
	memset ((char *) default_ifs, 0, 256 * sizeof (int));
	default_ifs ['\n'] = 1;
    }
}


void print_string (s)
    char *s;
{
    if (directprint) {
	show_message (s, 0);
	if (mode == MODE_INTERACT || mode == MODE_MINISHELL)
	    wrefresh (menuw);
    } else {
	int sz = strlen (s);
	/* The \0 is not put in the buffer, but we keep room for it. */
	if (print_buffer_pos+sz+1 > print_buffer_max) {
	    print_buffer_max += print_buffer_max + sz + 1;
	    print_buffer = mrealloc (print_buffer, print_buffer_max);
	}
	memcpy (print_buffer+print_buffer_pos, s, sz); /* without \0 */
	print_buffer_pos += sz;
    }
}





/***********************************************************************/
/***********************************************************************/
/***********************************************************************/

static List *build_list ();
static void exec_list ();
void exec_tree ();

static long op_add (i,j)
    long i, j;
{
    return i+j;
}

static long op_sub (i,j)
    long i, j;
{
    return i-j;
}

static long op_mul (i,j)
    long i, j;
{
    return i*j;
}

static long op_div (i,j)
    long i, j;
{
    if (j == 0) {
	set_fixed_error (ERR_EXEC, "division by zero");
	return 0;
    }
    return i/j;
}

static long op_lt (i,j)
    long i, j;
{
    return i < j; /* 1 or 0 */
}

static long op_le (i,j)
    long i, j;
{
    return i <= j; /* 1 or 0 */
}

static long op_gt (i,j)
    long i, j;
{
    return i > j; /* 1 or 0 */
}

static long op_ge (i,j)
    long i, j;
{
    return i >= j; /* 1 or 0 */
}

static long op_andand (i,j)
    long i, j;
{
    return i && j; /* 1 or 0 */
}

static long op_oror (i,j)
    long i, j;
{
    return i || j; /* 1 or 0 */
}

static long op_eq (s1,s2)
    char *s1, *s2;
{
    return STREQ(s1,s2); /* 1 or 0 */
}

static long op_neq (s1,s2)
    char *s1, *s2;
{
    return STREQ(s1,s2) == 0; /* 1 or 0 */
}


List *build_arith (tree)
    Node *tree;
{
    long (*op)();

    switch (tree->type) {
    case nAdd:
	op = op_add; goto do_op;
    case nSub:
	op = op_sub; goto do_op;
    case nMul:
	op = op_mul; goto do_op;
    case nDiv:
	op = op_div; goto do_op;
    case nLt:
	op = op_lt; goto do_op;
    case nLe:
	op = op_le; goto do_op;
    case nGt:
	op = op_gt; goto do_op;
    case nGe:
	op = op_ge; goto do_op;
    case nAndand:
	op = op_andand; goto do_op;
    case nOror:
	op = op_oror; goto do_op;
    do_op: {
	List *l1, *l2;
	long len1, len2;
	l1 = build_arith (tree->u[0].p);
	if (the_error) return NULL;
	if (l1 == NULL) {
	    set_fixed_error (ERR_EXEC,
			     "null argument to an arithmetic operation");
	    return NULL;
	}
	l2 = build_arith (tree->u[1].p);
	if (the_error) return NULL;
	if (l2 == NULL) {
	    set_fixed_error (ERR_EXEC,
			     "null argument to an arithmetic operation");
	    return NULL;
	}
	len1 = nbargs_list (l1);
	len2 = nbargs_list (l2);
	if (len1 == 1 || len2 == 1 || len1 == len2) {
	    int again;
	    List *ll; /* list we'll return */
	    if (len1 == 1)
		ll = l2;
	    else
		ll = l1;
	    do {
		long res1, res2;
		long res;
		char *str;
		if (   long_from_string (l1->arg, &res1)
		    || long_from_string (l2->arg, &res2)) {
		    set_fixed_error (ERR_EXEC, "operand is not a number");
		    return NULL;
		}
		res = (*op) (res1, res2);
		if (the_error) return NULL;
		str = sstring_from_long (res);
		if (len1 == 1) {
		    l2->arg = str;
		    l2 = l2->next;
		    again = (l2 != NULL);
		} else {
		    l1->arg = str;
		    l1 = l1->next;
		    if (len2 != 1)
			l2 = l2->next;
		    again = (l1 != NULL);
		}
	    } while (again);
	    return ll;
	}
	set_fixed_error (ERR_EXEC,
			 "lists of uneven size in arithmetic operation");
	return NULL;
    }
    case nEq:
	op = op_eq; goto do_strop;
    case nNeq:
	op = op_neq; goto do_strop;
    do_strop: {
	List *l1, *l2;
	long len1, len2;
	l1 = build_arith (tree->u[0].p);
	if (the_error) return NULL;
	if (l1 == NULL) {
	    set_fixed_error (ERR_EXEC, "null argument to a string operation");
	    return NULL;
	}
	l2 = build_arith (tree->u[1].p);
	if (the_error) return NULL;
	if (l2 == NULL) {
	    set_fixed_error (ERR_EXEC, "null argument to a string operation");
	    return NULL;
	}
	len1 = nbargs_list (l1);
	len2 = nbargs_list (l2);
	if (len1 == 1 || len2 == 1 || len1 == len2) {
	    int again;
	    List *ll; /* list we'll return */
	    if (len1 == 1)
		ll = l2;
	    else
		ll = l1;
	    do {
		long res;
		char *str;
		res = (*op) (l1->arg, l2->arg);
		if (the_error) return NULL;
		str = sstring_from_long (res);
		if (len1 == 1) {
		    l2->arg = str;
		    l2 = l2->next;
		    again = (l2 != NULL);
		} else {
		    l1->arg = str;
		    l1 = l1->next;
		    if (len2 != 1)
			l2 = l2->next;
		    again = (l1 != NULL);
		}
	    } while (again);
	    return ll;
	}
	set_fixed_error (ERR_EXEC, "lists of uneven size in string operation");
	return NULL;
    }
    case nArgs:
    case nArith:
    case nWord:
    case nVar:
    case nVarsub:
    case nCount:
    case nFlat:
    case nBackq:
    case nConcat: {
	return build_list (tree);
    }
    default:
	fprintf (stderr, "build_arith: forgotten or bad type %d\n",tree->type);
	return NULL;
    }
}


static List *build_list (tree)	      /* everything is in built in the stack */
    Node *tree;
{
    if (tree == NULL)
	return NULL;
    switch (tree->type) {
    case nArgs: {
	List *l1, *l2;
	l1 = build_list (tree->u[0].p);
	if (the_error) return NULL;
	l2 = build_list (tree->u[1].p);
	if (the_error) return NULL;
	return append_lists (l1,l2);
    }
    case nWord: {
	List *l = snew_List ();
	l->next = NULL;
	l->arg = tree->u[0].s;		    /* don't copy, this won't change */
	return l;
    }
    case nBackq: {
	int old_print_buffer_pos = print_buffer_pos;
	int olddirectprint = directprint;
	int sz;
	List *l;

	if (tree->u[0].p != NULL) {
	    l = build_list (tree->u[0].p);
	    if (the_error) return NULL;
	}
	directprint = 0;
	exec_tree (tree->u[1].p);
	directprint = olddirectprint;
	sz = print_buffer_pos - old_print_buffer_pos;
	print_buffer_pos = old_print_buffer_pos;
	if (the_error) return NULL;

	{
	    int *ifs; /* int is faster */
	    List *rl;
	    List **lp = &rl;
	    long i;
	    unsigned char *arg =
		(unsigned char *) print_buffer + print_buffer_pos;
	    arg[sz] = '\0';	 /* print_string ensures that we can do this */

	    if (tree->u[0].p == NULL) {
		while (1) {
		    List *al;
		    unsigned char c;
		    unsigned char *ptr;
		    ptr = arg;
		    while ((c = *arg) != '\n' && c != '\0')
			arg++;
		    i = arg - ptr;
		    al = snew_List ();
		    *lp = al;
		    al->arg = smalloc (command_stack, i+1);
		    if (i)
			memcpy (al->arg, ptr, i);
		    (al->arg)[i] = '\0';
		    lp = &(al->next);
		    if (c != '\0') /* skip a \n */
			c = *++arg;
		    if (c == '\0') {
			*lp = NULL;
			return rl;
		    }
		}
	    } else {
		List *ll;
		ifs = special_ifs;
		memset ((char *) ifs, 0, 256 * sizeof (int));
		for (ll = l; ll != NULL; ll = ll->next) {
		    unsigned char *ptr = (unsigned char *) ll->arg;
		    while (*ptr)
			ifs[*ptr++] = 1;		   /* *ptr is not \0 */
		}
		while (1) {
		    List *al;
		    unsigned char c;
		    unsigned char *ptr;
		    while (ifs[*arg])
			arg++;
		    if (*arg == 0) {
			*lp = NULL;
			return rl;
		    }
		    ptr = arg;
		    do {
			arg++;
		    } while (!ifs[c = *arg] && c != '\0');
		    i = arg - ptr;
		    al = snew_List ();
		    *lp = al;
		    al->arg = smalloc (command_stack, i+1);
		    memcpy (al->arg, ptr, i);
		    (al->arg)[i] = '\0';
		    lp = &(al->next);
		}
	    }
	}
    }
    case nConcat: {
	List *l1, *l2;
	l1 = build_list (tree->u[0].p);
	if (the_error) return NULL;
	l2 = build_list (tree->u[1].p);
	if (the_error) return NULL;
	return sconcat_lists (l1,l2); /* may raise an error */
    }
    case nArith: {
	return build_arith (tree->u[0].p);
    }
    case nVar:
    case nVarsub:
    case nCount:
    case nFlat: {
	List *var;
	List *l = build_list (tree->u[0].p);
	if (the_error) return NULL;
	if (l == NULL) {
	    set_fixed_error (ERR_EXEC, "null variable name");
	    return NULL;
	}
	if (l->next != NULL) {
	    set_fixed_error (ERR_EXEC, "multi-word variable name");
	    return NULL;
	}
	if (l->arg[0] == '\0') {
	    set_fixed_error (ERR_EXEC, "zero-length variable name");
	    return NULL;
	}
	var = fetch_variable (l->arg);
	/* BEWARE, var may be in permanent memory, so
	 *  - don't free it
	 *  - copy the args onto stack if needed
	 */
	switch (tree->type) {
	case nVar:
	    return scopy_of_list (var);				 /* in stack */
	case nVarsub: {
	    int n;
	    List *subs;
	    List *res;
	    List **last = &res;

	    subs = build_list (tree->u[1].p);
	    if (the_error) return NULL;
	    n = nbargs_list (var);
	    for (; subs != NULL; subs = subs->next) {
		long i = ulong_from_string (subs->arg);
		if (i < 0) {
		    set_error_list (subs->arg, NULL, "bad subscript");
		    return NULL;
		}
		if (i < n) {			    /* subscript begins at 0 */
		    List *sub = var;
		    List *cur = snew_List ();
		    while (i--)
			sub = sub->next;
		    cur->arg = sstrdup (command_stack, sub->arg);
		    *last = cur;
		    last = &(cur->next);
		}
	    }
	    *last = NULL;
	    return res;
	}
	case nCount: {
	    long i = nbargs_list (var);
	    List *l = snew_List ();
	    l->next = NULL;
	    l->arg = sstring_from_long (i);			 /* in stack */
	    return l;
	}
	case nFlat: {
	    int i;
	    List *l1, *l2;
	    char *w, *ww;

	    i = 0;
	    for (l1=var; l1 != NULL; l1 = l1->next)
		i += strlen (l1->arg) + 1;
	    w = ww = smalloc (command_stack, i);		 /* in stack */
	    for (; var != NULL; var = var->next) {
		int sz = strlen (var->arg);
		memcpy (ww, var->arg, sz);
		ww += sz;
		*ww++ = ' ';
	    }
	    ww[-1] = '\0';
	    l2 = snew_List ();
	    l2->next = NULL;
	    l2->arg = w;
	    return l2;
	}
	default: /* just to make gcc -Wall happy */
	    ;
	}
	/*NOTREACHED*/
    }
    default:
	fprintf (stderr, "build_list: forgotten or bad type %d\n", tree->type);
	return NULL;
    }
    /*NOTREACHED*/
}


static void exec_list (l)			   /* Doesn't free the list. */
    List *l;
{
    Fun *fun;

    if (l == NULL)
	return;
    fun = get_function (l->arg);
    if (fun != NULL) {
	int olduse = fun->inuse;	   /* You've heard about recursion ? */
	fun->inuse = 1;
	if (fun->builtin_fun != NULL) {
	    (*(fun->builtin_fun)) (l->next);
	} else {
	    Stack *old_stack = command_stack;
	    command_stack = push_stack (command_stack);
	    /* First arg will be $0. */
	    set_variable ("*", l->next, 1); /* push */
	    exec_tree (fun->tree);
	    delete_variable ("*");
	    command_stack = pop_stack (old_stack);
	    if (the_error == ERR_RETURN || the_error == ERR_BREAK)
		the_error = ERR_NONE;
	}
	fun->inuse = olduse;
	return;
    } else {
	set_error_list (l->arg, NULL, "not a function");
	return;
    }
}


#define EXEC_TREE(t) {tree=t; goto top;}
void exec_tree (tree)
    Node *tree;
{
 top:
    if (sigint) {
	sigint = 0;
	set_fixed_error (ERR_INTR, "Interrupted");
	return;
    }

    if (tree == NULL)
	return;

    switch (tree->type) {
    case nCmd:
	exec_tree (tree->u[0].p);
	if (the_error) return;
	EXEC_TREE (tree->u[1].p);
	/*NOTREACHED*/
    case nIf: {
	List *l;
	Node *n;
	Stack *old_stack = command_stack;
	command_stack = push_stack (command_stack);
	l = build_list (tree->u[0].p);
	if (the_error) return;
	if (l == NULL
	    || (l->arg[0] == '\0')
	    || (l->arg[0] == '0' && l->arg[1] == '\0'))
	    n = tree->u[2].p; /* else */
	else
	    n = tree->u[1].p; /* then */
	exec_tree (n);
	command_stack = pop_stack (old_stack);
	return;
    }
    case nFor: {
	List *forlist;
	List *l;
	char *arg;
	int push;
	List *varlist;
	Stack *old_stack = command_stack;
	command_stack = push_stack (command_stack);
	l = build_list (tree->u[0].p);
	if (the_error) return;
	if (l == NULL) {
	    set_fixed_error (ERR_EXEC, "null variable name");
	    return;
	}
	if (l->next != NULL) {
	    set_fixed_error (ERR_EXEC, "multi-word variable name");
	    return;
	}
	if (l->arg[0] == '\0') {
	    set_fixed_error (ERR_EXEC, "zero-length variable name");
	    return;
	}
	arg = l->arg;
	forlist = build_list (tree->u[1].p);
	if (the_error) return;
	push = 1;
	varlist = snew_List ();
	varlist->next = NULL;
	for (; !the_error && forlist != NULL; forlist = forlist->next) {
	    varlist->arg = forlist->arg;
	    store_variable (arg, varlist, push);
	    push = 0;
	    exec_tree (tree->u[2].p);
	}
	if (push == 0)
	    delete_variable (arg);
	command_stack = pop_stack (old_stack);
	return;
    }
    case nWhile: {
	List *l;
	Stack *old_stack;

	while (1) {
	    old_stack = command_stack;
	    command_stack = push_stack (command_stack);
	    l = build_list (tree->u[0].p);
	    if (the_error) return;
	    if (l == NULL
		|| (l->next == NULL
		    && (l->arg[0] == '\0'
			|| (l->arg[0] == '0' && l->arg[1] == '\0')))) {
		command_stack = pop_stack (old_stack);
		break;
	    }
	    exec_tree (tree->u[1].p);
	    command_stack = pop_stack (old_stack);
	}
	return;
    }
    case nFn: {
	List *l;
	Stack *old_stack = command_stack;
	command_stack = push_stack (command_stack);
	l = build_list (tree->u[0].p);
	if (the_error) return;
	if (l == NULL) {
	    set_fixed_error (ERR_EXEC, "null function name");
	    return;
	}
	if (l->next != NULL) {
	    set_fixed_error (ERR_EXEC, "multi-word function name");
	    return;
	}
	if (l->arg[0] == '\0') {
	    set_fixed_error (ERR_EXEC, "zero-length function name");
	    return;
	}
	if (tree->u[1].p == NULL)
	    remove_function (l->arg);
	else
	    add_function (l->arg, NULL, copy_tree (tree->u[1].p));
	command_stack = pop_stack (old_stack);
	return;
    }
    case nMap: {
	extern int special_cmd_map ();
	List *l;
	Stack *old_stack = command_stack;
	command_stack = push_stack (command_stack);
	l = build_list (tree->u[0].p);
	if (the_error) return;
	special_cmd_map (l, tree->u[1].p);	     /* has to copy the tree */
	command_stack = pop_stack (old_stack);
	return;
    }
    case nSetvar: {
	List *l1, *l2;
	Stack *old_stack = command_stack;

	command_stack = push_stack (command_stack);
	l1 = build_list (tree->u[0].p);
	if (the_error) return;
	if (l1 == NULL) {
	    set_fixed_error (ERR_EXEC, "null variable name");
	    return;
	}
	if (l1->next != NULL) {
	    set_fixed_error (ERR_EXEC, "multi-word variable name");
	    return;
	}
	if (l1->arg[0] == '\0') {
	    set_fixed_error (ERR_EXEC, "zero-length variable name");
	    return;
	}
	l2 = build_list (tree->u[1].p);
	if (the_error) return;
	store_variable (l1->arg, l2, 0); /* don't push */
	command_stack = pop_stack (old_stack);
	return;
    }
    case nLocal: {
	List *l, *ll, *l2;
	Stack *old_stack = command_stack;

	if (!inbraces) {
	    set_fixed_error (ERR_EXEC, "local: not in braces");
	    return;
	}
	command_stack = push_stack (command_stack);
	l = build_list (tree->u[0].p);
	if (the_error) return;
	if (l == NULL) {
	    set_fixed_error (ERR_EXEC, "null variable name");
	    return;
	}
	if (l->next != NULL) {
	    set_fixed_error (ERR_EXEC, "multi-word variable name");
	    return;
	}
	if (l->arg[0] == '\0') {
	    set_fixed_error (ERR_EXEC, "zero-length variable name");
	    return;
	}
	l2 = build_list (tree->u[1].p);
	if (the_error) return;
	ll = new_List (); /* permanent mem */
	ll->arg = mstrdup (l->arg);
	ll->next = local_variables;
	local_variables = ll;
	set_variable (l->arg, l2, 1); /* push (raw) */
	command_stack = pop_stack (old_stack);
	return;
    }
    case nReturn: {
	Stack *old_stack = command_stack;
	List *l, *ll;
	l = build_list (tree->u[0].p);
	if (the_error) return;
	if (l != NULL) {
	    long i = 0;
	    char *ptr, *string;
	    for (ll=l; ll != NULL; ll = ll->next)
		i += strlen (ll->arg) + 1;
	    ptr = string = smalloc (command_stack, i);
	    for (ll=l; ll != NULL; ll = ll->next) {
		int sz = strlen (ll->arg);
		memcpy (ptr, ll->arg, sz);
		ptr += sz;
		*ptr++ = '\n';
	    }
	    ptr [-1] = '\0';
	    print_string (string);
	}
	command_stack = pop_stack (old_stack);
	set_fixed_error (ERR_RETURN, "return: not in a function");
	return;
    }
    case nBrace: {
	List *old_local = local_variables;
	int old_inbraces = inbraces;
	inbraces = 1;
	local_variables = NULL;
	exec_tree (tree->u[0].p);
	while (local_variables != NULL) {
	    List *next = local_variables->next;
	    delete_variable (local_variables->arg);
	    mfree (local_variables->arg); /* permanent mem */
	    free_List (local_variables);
	    local_variables = next;
	}
	local_variables = old_local;
	inbraces = old_inbraces;
	if (the_error == ERR_BREAK && !--break_num)
	    the_error = ERR_NONE;
	return;
    }
    case nBreak: {
	Stack *old_stack = command_stack;
	List *l;
	l = build_list (tree->u[0].p);
	if (the_error) return;
	if (l == NULL)
	    break_num = 1;
	else if (l->next != NULL) {
	    set_fixed_error (ERR_EXEC, "break: too many arguments");
	    return;
	} else {
	    long res = ulong_from_string (l->arg);
	    if (res <= 0) {
		set_fixed_error (ERR_EXEC, "break: bad argument");
		return;
	    }
	    break_num = res;
	}
	command_stack = pop_stack (old_stack);
	set_fixed_error (ERR_BREAK, "break: not in braces");
	return;
    }
    case nVoid: {
	int old_print_buffer_pos = print_buffer_pos;
	int olddirectprint = directprint;
	directprint = 0;
	exec_tree (tree->u[0].p);
	directprint = olddirectprint;
	print_buffer_pos = old_print_buffer_pos;
	break;
    }
    case nCatch: {
	int old_print_buffer_pos = print_buffer_pos;
	int olddirectprint = directprint;
	directprint = 0;
	exec_tree (tree->u[0].p);
	directprint = olddirectprint;
	print_buffer_pos = old_print_buffer_pos;
	if (   the_error == ERR_LEX
	    || the_error == ERR_PARSE
	    || the_error == ERR_EXEC
	    || the_error == ERR_RAISE) {
	    print_string (the_error_message);
	    no_error ();
	}
	return;
    }
    case nArith:
    case nArgs:
    case nWord:
    case nVar:
    case nVarsub:
    case nCount:
    case nFlat:
    case nBackq:
    case nConcat: {
	List *l = build_list (tree);
	if (the_error) return;
	exec_list (l);
	return;
    }
    default:
	fprintf (stderr, "exec_tree: forgotten or bad type %d\n", tree->type);
	return;
    }
}



void exec_string (s)
    char *s;
{
    Node *tree;

    /* Assume the_error = 0 */
    tree = parse_string (s);
    if (the_error) return; /* no tree */
    exec_tree (tree);
    free_tree (tree);
}


void exec_file (file)
    FILE *file;
{
    parse_file_begin (file);
    do {
	Node *tree = parse_file_next ();
	if (the_error) break;
	exec_tree (tree);
	free_tree (tree);
    } while (!l_stream->eofmet && !the_error);
    parse_file_end ();
}
