/* funs.c */

#include "mem.h"
#include "tree.h"
#include "hash.h"
#include "builtin.h"
#include "builtin_interf.h"
#include "funs.h"




static Hashtable *function_table;


typedef struct Fun_queue Fun_queue;
struct Fun_queue {
    struct Fun_queue *next;
    Fun *fun;
};

static Fun_queue *pending_fun = NULL;

typedef struct {
    char *name;
    int (*builtin_fun) ();
} Assoc;

static Assoc orig_assocs [] =
{
    {"nop",		cmd_nop		},
    {"eval",		cmd_eval	},
    {"echo",		cmd_echo	},
    {"exec",		cmd_exec	},
    {"pipe",		cmd_pipe	},
    {"source",		cmd_source	},
    {"cd",		cmd_cd		},
    {"explode",		cmd_explode	},
    {"implode",		cmd_implode	},
    {"shift",		cmd_shift	},
    {"interval",	cmd_interval	},
    {"substring",	cmd_substring	},
    {"asc",		cmd_asc		},
    {"chr",		cmd_chr		},
    {"rnd",		cmd_rnd		},
    {"strlen",		cmd_strlen	},
    {"listlen",		cmd_listlen	},
    {"raise",		cmd_raise	},

    {"exit",		cmd_exit	},
    {"up",		cmd_up		},
    {"down",		cmd_dn		},
    {"left",		cmd_lt		},
    {"right",		cmd_rt		},
    {"pgup",		cmd_pgup	},
    {"pgdown",		cmd_pgdn	},
    {"pgleft",		cmd_pglt	},
    {"pgright",		cmd_pgrt	},
    {"scrollup",	cmd_scrollup	},
    {"scrolldown",	cmd_scrolldn	},
    {"scrollleft",	cmd_scrolllt	},
    {"scrollright",	cmd_scrollrt	},
    {"maxup",		cmd_maxup	},
    {"maxdown",		cmd_maxdn	},
    {"maxleft",		cmd_maxlt	},
    {"maxright",	cmd_maxrt	},
    {"ret",		cmd_return	},
    {"tab",		cmd_tab		},
    {"insert",		cmd_insert	},
    {"paste",		cmd_paste	},
    {"backspace",	cmd_backspace	},
    {"grab",		cmd_grab	},
    {"putchar",		cmd_putchar	},
    {"getchar",		cmd_getchar	},
    {"input",		cmd_input	},
    {"message",		cmd_message	},
    {"cls",		cmd_cls		},
    {"redraw",		cmd_redraw	},
    {"display",		cmd_display	},
    {"flush",		cmd_flush	},
    {"find",		cmd_find	},
    {"dumpkeymap",	cmd_dumpkeymap	},
    {"stats",		cmd_stats	},
    {"help",		cmd_help	},
    {"mapping",		cmd_mapping	},
    {"goto",		cmd_goto	},
    {"filter",		cmd_filter	},
    {"warp",		cmd_warp	},
    {"toplevel",	cmd_toplevel	},

    { NULL,		NULL		},
};



static void do_free_fun (fun)			      /* do it even if inuse */
    Fun *fun;
{
    if (fun != NULL) {
	free_tree (fun->tree); /* even if NULL */
	mfree (fun);
    }
}


void free_pending_functions ()
{
    while (pending_fun != NULL) {
	Fun_queue *next = pending_fun->next;
	do_free_fun (pending_fun->fun);
	mfree (pending_fun);
	pending_fun = next;
    }
}


static void free_function (fun)
    Fun *fun;
{
    if (fun->inuse) {
	Fun_queue *fun_queue = New (Fun_queue);
	fun_queue->fun = fun;
	fun_queue->next = pending_fun;
	pending_fun = fun_queue;
    } else
	do_free_fun (fun);
}



int remove_function (name)		       /* 1:succeeded, 0:nonexistent */
    char *name;
{
    Fun *fun;
    Bucket *bucket;

    bucket = delete_hashtable (function_table, name);
    if (bucket != NULL) {
	fun = (Fun *) (bucket->data);
	free_function (fun);
	mfree (bucket);
	return 1;
    }
    return 0;
}



void add_function (name, builtin_fun, tree)
    char *name;
    int (*builtin_fun) ();
    Node *tree;
{
    Fun *fun;
    Bucket *bucket;

    fun = New (Fun);
    fun->inuse = 0;
    fun->builtin_fun = builtin_fun;
    fun->tree = tree;
    bucket = insert_hashtable (function_table, name, (char *) fun);
    if (bucket != NULL) {
	fun = (Fun *) (bucket->data);
	free_function (fun);
	mfree (bucket);
    }
}

Fun *get_function (name)
    char *name;
{
    Bucket *bucket;

    bucket = search_hashtable (function_table, name);
    if (bucket != NULL)
	return (Fun *) (bucket->data);
    return NULL;
}



void init_function_table ()
{
    static int initialized = 0;

    if (!initialized) {
	Assoc *assoc = orig_assocs;
	initialized = 1;
	function_table = new_hashtable (0);
	while (assoc->name != NULL) {
	    add_function (assoc->name, assoc->builtin_fun, NULL);
	    assoc++;
	}
    }
}

