/* hash.c */

#include <string.h>

#include "mem.h"
#include "hash.h"


/*
 * When returning a bucket to be freed (delete & insert), the user
 * has to de-allocate its data and the bucket itself,
 * but NOT the key (already done).
 */

/*
 * freelist for Bucket
 */

static Bucket *Bucket_freelist = NULL;


static Bucket *new_Bucket ()		     /* only free_Bucket is exported */
{
    Bucket *b;
    if (Bucket_freelist) {
	b = Bucket_freelist;
	Bucket_freelist = b->next;
    } else
	b = New (Bucket);
    return b;
}

void free_Bucket (b)
    Bucket *b;
{
    b->next = Bucket_freelist;
    Bucket_freelist = b;
}



Hashtable *new_hashtable (nb)
    int nb;
{
    Hashtable *hashtable;

    if (!nb)
	nb = HASH_DEFAULT_SIZE;
    hashtable = New (Hashtable);
    hashtable->nb_buckets = nb;
    hashtable->bucket_array = (Bucket *) mmalloc (nb * sizeof (Bucket));
    /* Hopefully, a null pointer is represented as 0, or this won't work. */
    memset (hashtable->bucket_array, 0, nb * sizeof (Bucket));
    /* for (i=0; i<nb; i++)
	hashtable->bucket_array[i].next = (Bucket *) NULL; */
    return hashtable;
}


static int hash (hashtable, key)		      /* silly hash function */
    Hashtable *hashtable;
    char *key;
{
    unsigned int i = 0;

    while (*key)
	i += *key++;
    return (i % hashtable->nb_buckets);
}


/*
 * Return the bucket if it exists.  It must not be freed !
 */

Bucket *search_hashtable (hashtable, key)
    Hashtable *hashtable;
    char *key;
{
    Bucket *bucket;

    bucket = hashtable->bucket_array [hash (hashtable, key)].next;
    while (bucket != NULL && !STREQ (key, bucket->key))
	bucket = bucket->next;
    return bucket;
}


/*
 * Delete and return the bucket if it existed.
 */

Bucket *delete_hashtable (hashtable, key)
    Hashtable *hashtable;
    char *key;
{
    Bucket **bucketp;
    Bucket *bucket;

    bucketp = &(hashtable->bucket_array [hash (hashtable, key)].next);
    while (1) {
	bucket = *bucketp;
	if (bucket == NULL)
	    return NULL;
	if (STREQ (key, bucket->key)) {
	    *bucketp = bucket->next;
	    mfree (bucket->key);			     /* free the key */
	    return bucket;
	}
	bucketp = &(bucket->next);
    }
}


/*
 * Insert and return the previous bucket if it existed.  The key is copied !
 */

Bucket *insert_hashtable (hashtable, key, data)
    Hashtable *hashtable;
    char *key;
    char *data;
{
    Bucket *new;
    Bucket *bucket;
    Bucket **bucketp;
    Bucket **bucketp_o;

    new = new_Bucket ();
    new->key = mstrdup (key);
    new->data = data;
    bucketp = &(hashtable->bucket_array [hash (hashtable, key)].next);
    bucketp_o = bucketp;
    while (1) {
	bucket = *bucketp;
	if (bucket == NULL) {
	    new->next = *bucketp_o;			     /* Insert first */
	    *bucketp_o = new;
	    return NULL;
	}
	if (STREQ (key, bucket->key)) {
	    new->next = bucket->next;
	    *bucketp = new;
	    mfree (bucket->key);			     /* free the key */
	    return bucket;
	}
	bucketp = &(bucket->next);
    }
}
    
