/* lex.c */

#include "tree.h"
#include "y.tab.h"
#include "mem.h"
#include "readline.h"
#include "error.h"
#include "lex.h"

#include "externs.h"



extern Stack *parse_stack;


static Stream *l_stream_stack = NULL;
Stream *l_stream = NULL;
static char *l_error_message = NULL;			    /* not allocated */


static void l_ungetchar (c)
    unsigned char c;
{
    if (l_stream->pushpos == MAXPUSH) {
	fprintf (stderr, "Too much l_ungetchar\n");
	return;
    }
    l_stream->pushed [l_stream->pushpos++] = c;
    if (c == '\n') {
	l_stream->lineno--;
	l_stream->charno = l_stream->prevlen;
    }
    l_stream->prevchar = c;
    l_stream->charno--;
}


static int l_getchar ()
{
    int c;

    if (l_stream->prevchar == '\n') {
	l_stream->prevlen = l_stream->charno;
	l_stream->lineno++;
	l_stream->charno = 0;
    }
    if (l_stream->pushpos)
	c = l_stream->pushed [--l_stream->pushpos];
    else
	c = l_stream->getter (l_stream);

    l_stream->charno++;
    if (c < 0 || c > 255) {
	c = 0;
	l_stream->prevchar = 0;
	l_ungetchar (0);
    }
    if (c == 0)
	l_stream->eofmet = 1;
    l_stream->prevchar = c;
    return c;
}




static int l_getter_file (stream)
    Stream *stream;
{
    int c;
    do {
	c = getc (stream->file);
    } while (c == 0);
    return c;
}


static int l_getter_string (stream)
    Stream *stream;
{
    return *(stream->string)++;
}


static int l_getter_readline (stream)
    Stream *stream;
{
    unsigned char *str;

    str = stream->string;
    if (str == NULL) {
	str = (unsigned char *) readline
	    (stream->prompt2 ? "   >> ":"mur > ");	       /* XXX change */
	if (str == NULL)
	    return 0; /* EOF */
	add_history (str);
	stream->string = stream->string_orig = str;
    }
    if (*str) {
	(stream->string)++;
	return *str;
    }
    mfree (stream->string_orig);
    stream->string = NULL;
    return '\n';
}


static void l_pushstream ()
{
    l_stream->next = l_stream_stack;
    l_stream_stack = l_stream;
    l_stream = New (Stream);
    l_stream->pushpos = 0;
    l_stream->lineno = 0;
    l_stream->charno = 0;
    l_stream->prevchar = '\n';
    l_stream->prompt2 = 0;
    l_stream->eofmet = 0;
}

void l_popstream ()
{
    switch (l_stream->type) {
    case 0: /* file */
	if (l_stream->file != NULL)
	    fclose (l_stream->file);
	break;
    case 1: /* string */
	break;
    case 2: /* readline */
	if (l_stream->string != NULL)
	    mfree (l_stream->string_orig);
	break;
    }
    mfree (l_stream);
    l_stream = l_stream_stack;
    l_stream_stack = l_stream->next;
}

void l_setinput_file (file)		    /* file is closed by l_popstream */
    FILE *file;
{
    l_pushstream ();
    l_stream->type = 0;
    l_stream->file = file;
    l_stream->getter = l_getter_file;
}

void l_setinput_string (string)			/* string is not deallocated */
    char *string;
{
    l_pushstream ();
    l_stream->type = 1;
    l_stream->string = (unsigned char *) string;
    l_stream->getter = l_getter_string;
}


void l_setinput_readline ()
{
    l_pushstream ();
    l_stream->type = 2;
    l_stream->string = NULL;
    l_stream->getter = l_getter_readline;
}


static int specialbeg[256] = {
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
 /*   ! " # $ % & ' ( ) * + , - . /  0 1 2 3 4 5 6 7 8 9 : ; < = > ? */
    1,1,1,1,1,0,1,1,1,1,1,1,0,1,0,1, 0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,
 /* @ A B C D E F G H I J K L M N O  P Q R S T U V W X Y Z [ \ ] ^ _ */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,
 /* ` a b c d e f g h i j k l m n o  p q r s t u v w x y z { | } ~ ^? */
    1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,1,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
};

static int specialmid[256] = {
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
 /*   ! " # $ % & ' ( ) * + , - . /  0 1 2 3 4 5 6 7 8 9 : ; < = > ? */
    1,1,1,1,1,0,0,1,1,1,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,
 /* @ A B C D E F G H I J K L M N O  P Q R S T U V W X Y Z [ \ ] ^ _ */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,
 /* ` a b c d e f g h i j k l m n o  p q r s t u v w x y z { | } ~ ^? */
    1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
};


static int buffer_max = 100;
static char *buffer = NULL;

enum wtype { NW, RW, KW }; /* not a word, a real word, a keyword */
static enum wtype prevw = NW;


void init_yylex ()
{
    init_readline ();
    if (l_stream == NULL) {
	l_stream = New (Stream);
	l_stream->type = -1;
    }
    if (buffer == NULL)
	buffer = mmalloc (buffer_max);
}


void reset_yylex ()
{
    prevw = NW;
}

int real_yylex ()
{
#define CHK_CONCAT {if (prevw != NW) { prevw=NW; l_ungetchar(c); return '^'; }}
    int c;
    int i;

 top:
    while ((c = l_getchar ()) == ' ' || c == '\t')
	prevw = NW;
    l_stream->prompt2 = 1;
    if (c == 0)
	return END;
    if (!specialbeg [(unsigned char) c]) {			     /* word */
	CHK_CONCAT;
    read:
	i = 0;
	while (1) {
	    do {
		buffer[i++] = c;
		if (i >= buffer_max)
		    buffer = mrealloc (buffer, buffer_max *= 2);
	    } while ((c = l_getchar ()) != 0
		     && !specialmid [(unsigned char) c]);

	    if (c == '\\') {
		if ((c = l_getchar ()) == '\n') {      /* \ at EOL in a word */
		    c = l_getchar ();
		    if (!specialmid [(unsigned char) c])
			continue;
		} else if (c != '\0')
		    continue;
	    }
	    break;
	}
	l_ungetchar (c);
	buffer[i] = '\0';
	prevw = KW;
	if (i == 2) {
	    if (buffer[0] == 'i' && buffer[1] == 'f') return IF;
	}
	if (STREQ (buffer, "map")) return MAP;
	if (STREQ (buffer, "fun")) return FUN;
	if (STREQ (buffer, "for")) return FOR;
	if (STREQ (buffer, "then")) return THEN;
	if (STREQ (buffer, "else")) return ELSE;
	if (STREQ (buffer, "void")) return VOID;
	if (STREQ (buffer, "catch")) return CATCH;
	if (STREQ (buffer, "while")) return WHILE;
	if (STREQ (buffer, "local")) return LOCAL;
	if (STREQ (buffer, "return")) return RETURN;
	if (STREQ (buffer, "break")) return BREAK;
	prevw = RW;
	yylval.word = sstrdup (parse_stack, buffer);
	return WORD;
    }
    if (   c == '`'
	|| c == '$'
	|| c == '\''
	|| c == '"'
	|| c == '[') {
	CHK_CONCAT;
    }
    switch (c) {
    case '\0':
	return END;
    case '!': {
	c = l_getchar ();
	if (c == '=') {
	    prevw = NW;
	    return NEQ;
	}
	l_ungetchar (c);
	c = '!';
	goto read;
    }
    case '\'':
    case '"': {
	int endc = c;
	prevw = RW;
	i = 0;
	while ((c = l_getchar ()) != endc) {
	    buffer[i++] = c;
	    if (c == '\0') {
		l_error_message = "EOF in quoted string";
		return HUH;
	    }
	    if (i >= buffer_max)
		buffer = mrealloc (buffer, buffer_max *= 2);
	}
	buffer[i] = '\0';
	yylval.word = sstrdup (parse_stack, buffer);
	return WORD;
    }
    case '$':
	prevw = NW;
	c = l_getchar ();
	if (c == '#')
	    return COUNT;
	if (c == '^')
	    return FLAT;
	l_ungetchar (c);
	return '$';
    case '(':   /*')'*/
	if (prevw == RW)
	    c = VARSUB;
	prevw = NW;
	return c;
    case '&': case '|': {
	int cc = c;
	c = l_getchar ();
	if (c == cc) {
	    prevw = NW;
	    if (c == '&') return ANDAND; else return OROR;
	}
	l_ungetchar (c);
	c = cc;
	goto read;
    }
    case '#':
	while ((c = l_getchar ()) != '\n')
	    if (c == 0)
		return END;
	/*FALLTHROUGH*/
    case '\n':
    case ';':
    case '^':
    case  /*'('*/  ')':
    case '{': case '}':
    case '[': case ']':
    case '`':
	prevw = NW;
	return c;
    case '\\':
	if ((c = l_getchar ()) == '\n')
	    goto top; /* ignore \ at EOL */
	l_ungetchar (c);
	c = '\\';
	CHK_CONCAT;
	c = l_getchar ();
	if (c != 0)
	    goto read;
	return END;
    case '+': case '-':
    case '*': case '/':
    case '=': {
	int cc = c;
	c = l_getchar ();
	l_ungetchar (c);
	if (c == ' ' || c == '\t' || c == '\n') {
	    prevw = NW;
	    return cc;
	}
	c = cc;
	goto read;
    }
    case '<': case '>': {
	int cc = c;
	prevw = NW;
	c = l_getchar ();
	if (c == '=')
	    if (cc == '<') return LE; else return GE;
	l_ungetchar (c);
	if (cc == '<') return LT; else return GT;
    }
    default:
	prevw = NW;
	return c; /* eat this, yacc... */
    }
}


int yylex ()
{
    int res = real_yylex ();

    /* printf ("\t\t\t\t\tyylex returned %d\n", res); */
    return res;
}




void set_prompt1 ()
{
    l_stream->prompt2 = 0;
}





void yyerror (msg)
    char *msg;
{
    if (l_error_message == NULL) {
	set_error (ERR_PARSE, msg);
    } else {
	set_error (ERR_LEX, l_error_message);
	l_error_message = NULL;
    }
}
