/* mem.c */

#include <stdio.h>

#include "mem.h"

#include "externs.h"


extern char *malloc ();
extern char *realloc ();
extern void free ();

char *mmalloc (size)
    long size;
{
    char *mem;

    mem = malloc (size);
    if (mem != NULL)
	return mem;
    fprintf (stderr, "Out of memory, couldn't allocate %d bytes\n", size);
    abort ();
    exit (255);
}


void mfree (mem)
    char *mem;
{
    free (mem);
}


char *mrealloc (mem, size)
    char *mem;
    long size;
{
    if (mem == NULL)
	mem = malloc (size);
    else
	mem = realloc (mem, size);
    if (mem != NULL)
	return mem;
    fprintf (stderr, "Out of memory, couldn't rallocate %d bytes\n", size);
    abort ();
    exit (255);
}


char *mstrdup (string)
    char *string;
{
    long size = strlen (string) + 1;
    char *mem = mmalloc (size);
    memcpy (mem, string, size);
    return mem;
}




/*
 * Allocation in stack.  Stack is the public type.
 */

#define SBLOCKSIZE 8192
#define SDELTA 64

static Stackblock *sblock_list;			      /* list of free blocks */


static Stackblock *stackblock_alloc (size)
    long size;
{
    Stackblock *sb;
    Stackblock **p = &sblock_list;
    for (; (sb = *p) != NULL; p = &(sb->next)) {
	if (size <= sb->size)
	    break;
    }
    if (sb != NULL) {
	*p = sb->next;
    } else {
	sb = New (Stackblock);
	sb->size = ((size+SBLOCKSIZE+SDELTA-1) & ~(SBLOCKSIZE-1)) - SDELTA;
	sb->mem = mmalloc (sb->size);
    }
    sb->next = NULL;
    return sb;
}

static void free_stackblock (stackblock)	/* free it and all the nexts */
    Stackblock *stackblock;
{
    Stackblock *sb;
    for (sb = stackblock; sb->next != NULL; sb = sb->next)
	;
    sb->next = sblock_list;
    sblock_list = stackblock;
}

char *smalloc (stack, size)
    Stack *stack;
    long size;
{
    long pos;
    Stackblock *stackblock = stack->block;
    size = (size+sizeof(long)-1) & ~(sizeof(long)-1);
    if (size + (pos = stack->pos) <= stackblock->size) {
	stack->pos += size;
	return stackblock->mem + pos;
    } else {
	Stackblock *new = stackblock_alloc (size);
	stackblock->next = new;
	stack->block = new;
	stack->pos = size;
	return new->mem;
    }
}


static Stack *Stack_freelist = NULL;

static Stack *new_stack ()
{
    Stack *stack;
    if (Stack_freelist != NULL) {
	stack = Stack_freelist;
	Stack_freelist = stack->next;
    } else
	stack = New (Stack);
    stack->next = NULL;
    return stack;
}

static void free_stack (stack)			/* free it and all the nexts */
    Stack *stack;
{
    Stack *s;
    for (s = stack; s->next != NULL; s = s->next)
	;
    s->next = Stack_freelist;
    Stack_freelist = stack;
}


/*
 * To save the state :
 * - Save the current stack somewhere.
 * - Pass it to push_stack, and it will return the new current stack.
 * To return to ANY previous state :
 * - Pass the corresponding saved stack to pop_stack, it will
 *   return the new current stack.
 */

Stack *push_stack (stack)
    Stack *stack;
{
    Stack *new = new_stack ();
    stack->next = new;
    new->block = stack->block;
    new->pos = stack->pos;
    return new;
}

Stack *pop_stack (old)
    Stack *old;
{
    if (old->block->next != NULL) {
	free_stackblock (old->block->next);
	old->block->next = NULL;
    }
    if (old->next != NULL) {
	free_stack (old->next);
	old->next = NULL;
    }
    return old;
}

/*
 * To create a new stack, call create_stack.  It returns the
 * current (empty) stack, which is also the stack to pass to destroy_stack
 * to destroy it completely.
 */

static Stackblock *nullstackblock_freelist = NULL;

Stack *create_stack ()
{
    Stack *stack = new_stack ();

    if (nullstackblock_freelist != NULL) {
	stack->block = nullstackblock_freelist;
	nullstackblock_freelist = nullstackblock_freelist->next;
    } else {
	stack->block = New (Stackblock);
	stack->block->size = 0;
    }
    stack->block->next = NULL;
    stack->next = NULL;
    stack->pos = 0;
    return stack;
}

void destroy_stack (stack)
    Stack *stack;
{
    pop_stack (stack);
    nullstackblock_freelist = stack->block;
    free_stack (stack);
}







char *sstrdup (stack, string)
    Stack *stack;
    char *string;
{
    int size = strlen (string) + 1;
    char *mem = smalloc (stack, size);
    memcpy (mem, string, size);
    return mem;
}

