/* scrio.c */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h> /* for FIONREAD */

#include "mur.h"
#include "commands.h"
#include "toplevel.h"
#include "error.h"
#include "screen.h"
#include "scrio.h"
#include "sysstuff.h" /* last */

#include "externs.h"


int cols = 80; /* we need default values */
int lines = 24;
int in_screen = 0;
int minishelllines = 5;	/* guaranteed at least 1 */
int newminilines = 5;
int is_xterm;
int want_mouse = 1;
int slow_mouse = 0;
int mouse_on = 0;

int flag_beep = 1; /* do we beep ? */




WINDOW *mainw, *menuw;

static int *keybuf = NULL; /* int */
static int keybufsize = 100;
static int keybuflen;




char *MY_DL, *MY_AL;
strmax termcap_db;
strmax termcap_data;



int sig_winch_pend = 0;
int sig_winch_ok = 0;
void (*sig_winch_routine) ();



int sigint = 0;

void sig_int ()
{
    sigint = 1;
    signal (SIGINT, sig_int);
}



static STRUCTTERMIOS original_tt;
static STRUCTTERMIOS getchar_tt;
static STRUCTTERMIOS running_tt;
static STRUCTTERMIOS incurses_tt;
static int original_flags;
static int getchar_flags;
static int running_flags;
static int incurses_flags;


void init_tty ()
{
    static int initialized = 0;

    if (initialized)
	return;
    initialized = 1;

    original_flags = fcntl (0, F_GETFL, 0);
    tcgetattr (0, &original_tt);

    getchar_flags = original_flags | O_NDELAY;
    getchar_tt = original_tt;
    getchar_tt.c_lflag &= ~(ISIG | ICANON | ECHO);
    
    running_flags = original_flags;
    running_tt = original_tt;
    running_tt.c_lflag |= ISIG;
    running_tt.c_lflag &= ~(ICANON | ECHO);
    running_tt.c_cc[VQUIT] = -1;
    running_tt.c_cc[VSUSP] = -1;
    running_tt.c_cc[VINTR] = 3;	/* ^C */
}

void stty_original ()
{
    fcntl (0, F_SETFL, original_flags);
    signal (SIGINT, SIG_IGN);
    tcsetattr (0, TCSANOW, &original_tt);
}

void stty_getchar ()
{
    fcntl (0, F_SETFL, getchar_flags);
    tcsetattr (0, TCSANOW, &getchar_tt);
}

void stty_running ()
{
    signal (SIGINT, sig_int);
    if (in_screen) {
	fcntl (0, F_SETFL, incurses_flags);
	tcsetattr (0, TCSANOW, &incurses_tt);
    } else {
	fcntl (0, F_SETFL, running_flags);
	tcsetattr (0, TCSANOW, &running_tt);
    }
}













static char mousechgbuf[] = "\033[?1000h";


void init_specialvars ()
{
    extern char *getenv ();
    char *xterm = getenv ("TERM");
    is_xterm = (xterm != NULL && !strncmp (xterm,"xterm",5));
}


void set_mouse_on ()
{
    if (is_xterm && !mouse_on) {
	mouse_on = 1;
	mousechgbuf[7] = 'h';
	write (1, mousechgbuf, 8);
    }
}

void set_mouse_off ()
{
    if (is_xterm && mouse_on) {
	mouse_on = 0;
	mousechgbuf[7] = 'l';
	write (1, mousechgbuf, 8);
    }
}



void get_scr_size ()
{
    lines = LINES-minishelllines;
    if (lines < 1) {
	lines = 1;
	minishelllines = LINES-lines;
	if (minishelllines < 1)
	    minishelllines = 1;
    }
    cols = COLS-1; /* 1 char on the right because curses is broken... */
    if (cols < 1)
	cols = 1;
}


static void signal_winch ();

void init_scr ()
{
    static int curses_initialized = 0;

    if (!in_screen)
    {
	in_screen = 1;

	fflush (stdout);
#ifdef NEWCURSES
	if (!curses_initialized) {
	    curses_initialized = 1;
#else
	    LINES = COLS = 0;
#endif
	    initscr ();
	    get_scr_size ();
	    mainw = subwin (stdscr, lines, cols, 0, 0);
	    if (mainw == (WINDOW *) ERR)
		mainw = subwin (stdscr, LINES, COLS, 0, 0);  /* if too small */
	    menuw = subwin (stdscr, minishelllines, cols, lines, 0);
	    if (menuw == (WINDOW *) ERR)
		menuw = subwin (stdscr, LINES, COLS, 0, 0);  /* if too small */
	    leaveok (mainw, FALSE);
	    leaveok (menuw, FALSE);
#ifdef NEWCURSES
	} else {
	    wrefresh (menuw);
	    wrefresh (mainw);
	}
#else
	reprint_screen (1);
	wrefresh (menuw);
	wrefresh (mainw);
#endif

	keybuflen = 0;

	incurses_flags = fcntl (0, F_GETFL, 0);
	tcgetattr (0, &incurses_tt);
	incurses_tt.c_lflag |= ISIG;
	incurses_tt.c_lflag &= ~(ICANON | ECHO);
	incurses_tt.c_cc[VQUIT] = -1;
	incurses_tt.c_cc[VSUSP] = -1;
	incurses_tt.c_cc[VINTR] = 3;	/* ^C */

	signal (SIGWINCH, signal_winch);

    }
}


void exit_scr ()
{
    if (in_screen) {
	in_screen = 0;
#ifndef NEWCURSES
	delwin (mainw);
	delwin (menuw);
#endif
	endwin ();
    }
}





void beep ()
{
    static unsigned char c = 7;
    if (flag_beep)
	write (1, &c, 1);
}



void init_keys ()
{
    if (keybuf == NULL)
	keybuf = (int *) mmalloc (keybufsize * sizeof (int));
    keybuflen = 0;
}




static int read_key ()		      /* Must be called with an empty keybuf */
{
#define TMPKEYSIZE 1000
    static unsigned char tmpkey [TMPKEYSIZE];
    /*
     * receivestate : 1 is after Esc, 2 after Esc [, 3 after Esc [ M,
     * 4 after Esc [ M c0, 5 after Esc [ M c0 c1
     */
    static int receivestate = 0;
    static int mouse_key_but;
    static int mouse_key_x;
    int i;
    int len;
    fd_set fdset;

 top:
    FD_ZERO (&fdset);
    FD_SET (0, &fdset);
    i = select (1, &fdset, NULL, NULL, NULL);
    if (i < 0)
	return -2; /* interrupted */
    len = read (0, tmpkey, TMPKEYSIZE);
    if (len <= 0) {
	if (len == 0 || errno != EINTR) {
	    return -1; /* eof / other error */
	} else {
	    return -2; /* interrupted */
	}
    }
    /* Here we assume that keybuflen = 0 */
    if (len + 10 >= keybufsize) {
	keybufsize += len + 20;
	keybuf = (int *) mrealloc (keybuf, keybufsize * sizeof (int));
    }
    if (is_xterm) {
	int lastpos = len+10;
	int pos = lastpos;
	for (i=0; i<len; i++) {
	    unsigned char c = tmpkey [i];
	    switch (receivestate) {
	    case 0:
		if (c == 0x1b)
		    receivestate = 1;
		else
		    keybuf [--pos] = c;
		break;
	    case 1:
		if (c == '[')
		    receivestate = 2;
		else {
		    keybuf [--pos] = 0x1b;
		    keybuf [--pos] = c;
		    receivestate = 0;
		}
		break;
	    case 2:
		if (c == 'M')
		    receivestate = 3;
		else {
		    keybuf [--pos] = 0x1b;
		    keybuf [--pos] = '[';
		    keybuf [--pos] = c;
		    receivestate = 0;
		}
		break;
	    case 3:
		mouse_key_but = c;
		receivestate = 4;
		break;
	    case 4:
		mouse_key_x = c;
		receivestate = 5;
		break;
	    case 5: {
		if ((mouse_key_but & 3) != 3) {	   /* if not release */
		    keybuf [--pos] =
			((mouse_key_but & 0x1f)		/* C/M/S/x/x */
			 | 0x100
			 | ((mouse_key_x - 0x21) << 24)
			 | ((c - 0x21) << 16));
		}
		receivestate = 0;
		break;
	    }
	    }
	}
	keybuflen = lastpos-pos;
	if (keybuflen != 0)
	    bcopy (keybuf+pos, keybuf, keybuflen * sizeof (int));
    } else {
	keybuflen = len;
	for (i=0; i<len; i++) {
	    keybuf [keybuflen-i-1] = tmpkey [i];
	}
    }

    if (keybuflen == 0) { /* if a prefix of Esc [ M x y z  or  release but */
	switch (receivestate) {
	case 5:
	    goto top;
/*	    keybuf [keybuflen++] = mouse_key_x;*/
	case 4:
	    goto top;
/*	    keybuf [keybuflen++] = mouse_key_but;*/
	case 3:
	    goto top;
/*	    keybuf [keybuflen++] = 'M';*/
	case 2:
	    if (slow_mouse) goto top;
	    keybuf [keybuflen++] = '[';
	case 1:
	    if (slow_mouse) goto top;
	    break;
	case 0:	/* if release of button */
	    goto top;
	}
	receivestate = 0;
	return 0x1b;
    } else
	return keybuf [--keybuflen];
}

int get_key ()
{
    int key;

    if (keybuflen)
	return keybuf [--keybuflen];
    stty_getchar ();
    key = read_key ();
    stty_running ();
    if (key == -1) { /* EOF */
	set_error (ERR_TOPLEVEL, "");
	newmode = MODE_EXIT;
    }
    return key;
}

void unget_key (key)
    int key;
{
    keybuf [keybuflen++] = key;
}


void clear_key_buffer ()
{
    keybuflen = 0;
}


int key_pressed ()
{
    int nb;

    if (keybuflen)
	return 1;
    stty_getchar ();
    ioctl (0, FIONREAD, &nb);
    stty_running ();
    return (nb > 0);
}






static void signal_winch ()
{
    sig_winch_pend = 1;
    if (sig_winch_ok) {
	(*sig_winch_routine) ();
	sig_winch_pend = 0;
    }
}

